/*

   TLICatty v3 - artificial futility
   ------------------------------
   Copyright (C) 2004 by Michal Zalewski <lcamtuf@coredump.cx>
   TLI Modifications by Michal Spadlinski <gim@skrzynka.pl>

   See README for more details.

*/

#define __EXTENSIONS__		/* Solaris setbuffer(),fdopen() */
#define __USE_BSD		/* Linux setbuffer() */
#define __USE_POSIX		/* Linux fdopen() */
#include <stdio.h>
#undef __USE_POSIX
#include <unistd.h>
#include <stdlib.h>
#define __EXTENSIONS__		/* Solaris bzero() */
#define __USE_XOPEN_EXTENDED	/* Linux bzero() */
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <sys/types.h>
#include <fcntl.h>

/* MY_STUFF_BEGIN_1 */
#include <time.h>
#define __USE_POSIX		/* Linux kill() */
#include <signal.h>
#define _XOPEN_SOURCE
#include <xti.h>

#include <netinet/in.h>		/* INADDR_ANY */
#include <sys/socket.h>		/* AF_INET */
#include <sys/stropts.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>


#define DEF_PORT 0xdeaf
#define DEF_QUEUE 5

#if !defined(FIONBIO)
#define GPID "%ld"
#else
#define GPID "%d"
#endif

#ifndef HL_AIX
#define DEF_T_OPEN_TCP_IPV4 "/dev/tcp"
#else
#define DEF_T_OPEN_TCP_IPV4 "/dev/xti/tcp"
#endif
/* MY_STUFF_END_1 */

#include "types.h"
#include "config.h"

#define fatal(x...) do { \
    fprintf(stderr,"[-] FATAL: "); \
    fprintf(stderr,x); \
    fprintf(stderr,"\n"); \
    exit(1); \
  } while (0)

#define R(x) (rand() % (x))


static _s32 kbf;

struct topic {
  _u8  *tw[MAXKEY+1], *pw[MAXPOP+1];
  _u32 fp;
  _u32 cnt, ok;
  _u32 score;
};

static struct topic top[MAXTOPICS];
static _s32 tcur = -1, wcnt;

struct phrase {
  _u8 t[MAXWORDS+2][MAXWLEN+1];
};

static struct phrase curp, workp, ctx[KEEP_CTX];

struct repeat {
  _u32 fpos;
  _u32 top;
};

static struct repeat rep[REPEAT_LOG];

struct pscore {
  _u32 fpos;
  _u32 score;
};

static struct pscore psc[MAXPHRASES];


static void index_kbase(void) {
  _u8 buf[LINEBUF];
  FILE* f;
  _u32 cnt_ent = 0;

  if ((kbf = open(KBASE,O_RDONLY)) < 0)
    fatal("Unable to open %s.",KBASE);

  f = fdopen(kbf,"r");

  while (fgets(buf,LINEBUF,f)) {
    if (buf[0] == '!') {
      _u8 *x, *y;
      _u32 i = 0;

      if (tcur >= 0 && !top[tcur].cnt) tcur--;

      if (++tcur >= MAXTOPICS) 
        fatal("Number of topics in database exceeds MAXTOPICS.");

      x = y = buf + 1;

      while ((x = strtok(y," "))) { 
        y = 0;
        top[tcur].tw[i++] = strdup(x);
        if (i >= MAXKEY)
          fatal("Some topics have more than MAXKEY keywords.");
      }

    } else if (buf[0] == '@') {
      _u8 *x, *y;
      _u32 i = 0;

      if (tcur < 0) fatal("Knowledge database does not start with '!'.");

      top[tcur].fp = ftell(f);
      x = y = buf + 1;

      while ((x = strtok(y," "))) { 
        y = 0;
        top[tcur].pw[i++] = strdup(x);
        if (i >= MAXPOP)
          fatal("Some topics have more than MAXPOP pop-words.");
      }

    } else {
      if (tcur < 0 || !top[tcur].fp) 
        fatal("Knowledge database does not start with '!' and '@'.");
      if (++top[tcur].cnt >= MAXPHRASES)
        fatal("Some topics have more than MAXPHRASES phrases.");
      top[tcur].ok++;
      cnt_ent++;
    }
  }

  kbf = dup(kbf);
  fclose(f);

  if (tcur >= 0 && !top[tcur].cnt) tcur--;
  if (tcur < 0) fatal("Empty database!");

  printf("Database indexed: %u topics, %u entries.\n", tcur + 1, cnt_ent);  

}


static void split_words(_u8* x,struct phrase* wk) {
  _u8 cw = 0, cp = 0;

  while (*x) {
    if (isalnum(*x)) {
      wk->t[cw][cp] = tolower(*x);
      if (cp < MAXWLEN) cp++;
    } else {
      if (cp) {
        wk->t[cw][cp] = 0;
        if (cw < MAXWORDS) cw++;
        cp = 0;
      }
    }
    x++;
  }

  wk->t[cw][0] = 0;
  wcnt = cw;

}


static _s32 find_best_topic(void) {
  _u32 b,c,d;
  _u8  *CW, *TW;

  _s32 tpos = -1;
  _u32 tval = 0;

  /* Lower topic scores */
  for (c=0;c<=tcur;c++) top[c].score >>= 3;

  b = 0;

  /* For every word of user input */
  while ((CW=curp.t[b])[0]) {

    /* For every topic */
    for (c=0;c<=tcur;c++) {

      /* Skip topics that have no spare lines */
      if (!top[c].ok) continue;

      /* The following is a bit sub-optimal... */

      d = 0;
      /* For every word of the topic */
      while ((TW=top[c].tw[d])) {

        if (!strcmp(CW,TW)) {
          _u32 z = 1, delta;
          _u8 *zTW, *zCW;

          delta = strlen(CW) * strlen(CW);

          /* A dirty hack to include run-ons */
          while ((zCW=curp.t[b+z])[0] && (zTW=top[c].tw[d+z]) &&
                 !strcmp(zCW,zTW)) { delta <<= 1; z++; }

          top[c].score += delta << 1;
          break;

        }

        d++;
      }
  
      d = 0;
      /* For every pop-word of the topic */
      while ((TW=top[c].pw[d])) {

        /* Run-ons are meaningless in pop words */
        if (!strcmp(CW,TW)) {
          top[c].score += (strlen(CW) * strlen(CW)) >> 1;
          break;
        }

        d++;
      }

#define R100() (85 + (rand() % 30))

      if (top[c].score * R100() > tval * 100) { 
        tval = top[c].score; tpos = c; 
      }
  
    }

    b++;

  }

  return tpos;

}


static _u32 find_best_phrase(_u32 topic,_u8 isaq) {
  FILE* f;
  _u8 buf[LINEBUF];
  _u32 pos = top[topic].fp, pnum = 0, tpos = 0, tscor = 0;

  f = fdopen(kbf,"r");
  fseek(f, pos, SEEK_SET);

  while (fgets(buf,LINEBUF,f)) {
    _u32 i,j,v;
    _u8 *CW, *WW, thisq = 0;

    if (buf[0] == '!') break;

    /* Show wits, do not be overly pedantic... */
    if (pnum && !R(2)) goto skip_me;

    /* Skip responses already used up... */
    for (i=0;i<REPEAT_LOG;i++)
      if (rep[i].fpos == pos) goto skip_me;

    if (strchr(buf,'?')) thisq = 1;

    split_words(buf,&workp);

    psc[pnum].score = 0;
    psc[pnum].fpos = pos;

    /* For every context line */
    for (v=0;v<KEEP_CTX;v++) {

      /* For every word of user input */
      j = 0;
      while ((CW=ctx[v].t[j])[0]) {

        i = 0;
        /* For every word of tested phrase */
        while ((WW=workp.t[i])[0]) {

          if (!strcmp(CW,WW)) {
            _u32 z = 1, delta;
            _u8 *zWW, *zCW;
  
            delta = strlen(WW) * strlen(WW);
  
            /* A dirty hack to include run-ons */
            while ((zWW=workp.t[i+z])[0] && (zCW=ctx[v].t[j+z])[0] &&
                   !strcmp(zCW,zWW)) { delta <<= 1; z++; }
  
            psc[pnum].score += delta << v;
            break;
  
          }
  
          i++;
        }
        j++;
      }

    }

    /* Penalty for answering with questions... */
    if (isaq == thisq) psc[pnum].score >>= 1;
  
    if (psc[pnum].score * R100() > tscor * 100) {
      tscor = psc[pnum].score; tpos = pnum;
    }

    pnum++;

skip_me:
    pos = ftell(f);

  }

  kbf = dup(kbf);
  fclose(f);

  if (!tscor) tpos = rand() % pnum;
  return psc[tpos].fpos;

}


static void display_phrase(int fd, _u8* prefix, _u32 poff) {
  FILE *f;
  char buf[LINEBUF], *x;

  f = fdopen(kbf,"r");
  fseek(f, poff, SEEK_SET);

  fgets(buf,LINEBUF,f);

  x = buf;
  while ((*x) && isspace((int)(*x))) x++;

  /*
  printf("%s: %s", prefix, x);
  */
	if (write(fd, x, strlen(x)) ==	strlen(x))
	{
		if (strlen(x)>60) {
			if (x[strlen(x)-1]=='\n') x[strlen(x)-1]=0;
	                x[55]=0;
        	        strcat(x, "[..]\n");
		}
		printf(" (G"GPID"+) n_reply send: %s", getpid(), x);
	} else {
		perror(" (G!) write()");
		printf(" (G"GPID"!) error in write in normal answer\n", getpid());
	}

  kbf = dup(kbf);
  fclose(f);

}


static _u8* excuse[MAXEXC];
static _u32 ecnt;

static void load_excuses(void) {
  FILE* f = fopen(EXCUSES,"r");
  _u8 buf[LINEBUF];

  if (!f) fatal("Cannot open %s.",EXCUSES);
  while (fgets(buf,LINEBUF,f)) {
    _u8* x = buf;
    while (*x && isspace(*x)) x++;
    while (strlen(x) > 0 && isspace(x[strlen(x)-1])) x[strlen(x)-1] = 0;
    if (!*x) continue;
    excuse[ecnt++] = strdup(x);
    if (ecnt >= MAXEXC) fatal("Number of excuses exceeds MAXEXC.");
  }

  fclose(f);

}


/* MY_STUFF_BEGIN_2 */
static void display_excuse(int fd, _u8* prefix) {
	int whichone = R(ecnt);
	char buf[LINEBUF];
	strncpy(buf, excuse[whichone], LINEBUF);
	strcat(buf, "\n");
	if (write(fd, buf, strlen(buf)) == strlen(buf)) {
		if (strlen(buf)>60) {
			strncpy(buf, excuse[whichone], 55);
			if (buf[strlen(buf)-1]=='\n')
				buf[strlen(buf)-1]=0;
			strcat(buf, "[..]\n");
		}
		printf(" (G"GPID"+) e_reply send: %s", getpid(), buf);
	} else {
		perror(" (G!) write()");
		printf(" (G"GPID"!) error in write in excuse\n", getpid());
	}
}
/* MY_STUFF_END_2 */


static void schedule_learn(void) {
  _u32 c = R(MAXKEY-1)+1;
  _s32 s = R(wcnt) - c;
  _u8 buf[(MAXKEY+4)*(MAXWLEN+4)+1];

  if (s < 0) s = 0;
  buf[0] = 0;

  while (c--) {
    if (!curp.t[s][0]) break;
    if (buf[0]) strcat(buf,"+");
    strcat(buf,curp.t[s]);
    s++;
  }

  if (buf[0]) { 
    if (chdir("cron/")) return;
    close(creat(buf,0644));
    chdir("..");
  }

}


static void store_pid(void) {
  FILE* x; 
  if (!(x=fopen(".cattypid","w"))) return; 
  fprintf(x,GPID"\n",getpid()); 
  fclose(x); 
}


/* MY_STUFF_BEGIN_3 */
extern int t_errno;
int accept_client(int listenfd, struct t_call *call, int rwflag)
{
	int newendpoint;

	if ((newendpoint = t_open(DEF_T_OPEN_TCP_IPV4, O_RDWR, NULL)) < 0)
	{
		t_error(" (G!) fatal! t_open()");
		return -1;
	}

	/* niech system nam przydzieli sokecik dla nowego klienta */
	if (t_bind(newendpoint, NULL, NULL) < 0)
	{
		t_error(" (G!) failed in t_bind()");
		t_close(newendpoint);
		return -1;
	}
	/* po wywolaniu socket jest w stanie T_UNBND i dopiero
	 * t_accept go tak naprawd podbinduje...
	 */

	if (t_accept(listenfd, newendpoint, call) < 0)
	{
		if (t_errno == TLOOK)
		{
			printf(" (G!) %d\n", t_look(listenfd));
			t_rcvdis(listenfd, NULL);
			t_unbind(newendpoint);
			t_close(newendpoint);
			return -1;
		}
	}

	if (rwflag)
	{
		if (ioctl(newendpoint, I_POP, NULL) < 0 ||
				ioctl(newendpoint, I_PUSH, "tirdwr") < 0)
		{
			t_snddis(newendpoint, NULL);
			t_unbind(newendpoint);
			t_close(newendpoint);
			return -1;
		}

	}
	return newendpoint;
}

#define MAXCHILD 5
int globfd = 0;
int dzieci[MAXCHILD];
void sigint_child_handler(int sig)
{
	printf(" (G-) in child, exiting...\n");
	t_unbind(globfd);
	t_close(globfd);
	exit(0);
}

int we_all_have_a_part_of_catty_in_ourself(int fd)
{
  	char buf[LINEBUF], temp[80];
	int n;

	globfd = fd;
	signal(SIGINT, sigint_child_handler);
	while (1) {
		if ((n=read(fd, buf, LINEBUF)) < 0) {
			perror(" (G!) read() error");
			printf(" (G"GPID"!) Closing connection\n", getpid());
		} else if (n == 0) {
			printf(" (G"GPID"+) Client exited!\n", getpid());
		} else {
			buf[n]=0;
			strcpy(temp, buf);
			if (strlen(buf)>60) {
				strncpy(temp, buf, 55);
				temp[55]=0;
				if (strchr(temp,'\n'))
					*strchr(temp, '\n') = 0;
				if (strchr(temp,'\r'))
					*strchr(temp, '\r') = 0;
				strcat(temp, "[..]\n");
			}
			printf(" (G"GPID"+) received: %s", getpid(), temp);
			/* Catty code goes here */
			_u8* x, isaq = 0;
			_s32 best, bpos;
			_u32 i;

			/*
			if (!(x=strchr(buf,':'))) continue;
			*x = 0; x++;
			*/
			x = buf;

			if (strchr(x,'?')) isaq = 1;

			split_words(x,&curp);
			if (!R(10) && wcnt) schedule_learn();
			best = find_best_topic();

			/* From time to time, be spontaneous */
			if ((best < 0 && !R(20)) || !R(90)) {
				best = rand() % (tcur+1);
				if (!top[best].ok) best=-1;
			}

			if (best < 0) display_excuse(fd, buf); 
			else {

				/* Save context */
				for (i=0;i<KEEP_CTX-1;i++) 
					memcpy(&ctx[i],&ctx[i+1],sizeof(struct phrase));
				memcpy(&ctx[KEEP_CTX-1],&curp,sizeof(struct phrase));

				bpos = find_best_phrase(best,isaq);
				display_phrase(fd, buf, bpos);

				/* Update repeat prevention data */
				top[best].ok--;

				if (rep[0].fpos) top[rep[0].top].ok++;

				for (i=0;i<REPEAT_LOG-1;i++) 
					memcpy(&rep[i],&rep[i+1],sizeof(struct repeat));

				rep[REPEAT_LOG-1].fpos = bpos;
				rep[REPEAT_LOG-1].top  = best;
			}
			continue;
		}
		t_rcvdis(fd, NULL);
		t_unbind(fd);
		t_close(fd);
		return 0;
	}
}

void sigchld_handler (int sig)
{
	int             pid, j;

	printf(" (G+) in sigchld routine waiting for children...\n");
	while ((pid = waitpid (-1, NULL, WNOHANG)) > 0)
	{
		for (j=0; j<MAXCHILD; j++)
			if (dzieci[j] == pid) {
				dzieci[j] = 0;
				printf(" (G+) Child exited: %d\n", pid);
				break;
			}
	}
	signal (SIGCHLD, sigchld_handler);
}
void sigint_handler(int sig)
{
	int i;
	printf("\n");
	for (i=0; i<MAXCHILD; i++)
		if (dzieci[i]) {
			kill(dzieci[i], SIGINT);
			waitpid(dzieci[i], NULL, 0);
		}

	if (t_unbind(globfd) < 0)
	{
		t_error(" (G!) unbind()");
	}
	t_close(globfd);
	printf(" *** SERVER SHUT DOWN ***\n");
	exit (0);
}

int prepare_tli(int *s, struct sockaddr_in *dst, 
		struct t_bind *b, struct t_call *c)
{
	if ((*s = t_open(DEF_T_OPEN_TCP_IPV4, O_RDWR, NULL)) < 0)
	{
		t_error(" (G!) fatal! t_open()");
		return 1;
	}
	puts(" (G+) endpoint opened");

	bzero((char *)dst, sizeof(struct sockaddr_in));
	dst->sin_family = AF_INET;
	dst->sin_addr.s_addr = htonl(INADDR_ANY);
	dst->sin_port = htons(DEF_PORT);

	/* poniewaz dostawca jest tcp to... */
	b->addr.maxlen = sizeof(struct sockaddr_in);
	b->addr.len = sizeof(struct sockaddr_in);
	b->addr.buf = (char *)dst;
	b->qlen = DEF_QUEUE;

	if (t_bind(*s, b, NULL) < 0)
	{
		t_error(" (G!) failed in t_bind()");
		t_close(*s);
		return 1;
	}
	puts(" (G+) endpoint binded");

	c->addr.maxlen = sizeof(struct sockaddr_in);
	c->addr.len = sizeof(struct sockaddr_in);
	c->addr.buf = (char *) dst;
	c->opt.maxlen = 0;
	c->opt.len = 0;
	c->udata.maxlen = 0;
	c->udata.len = 0;

	globfd = *s;
	signal(SIGINT, sigint_handler);
	signal (SIGCHLD, sigchld_handler);
	return 0;
}

int main(int argc,char** argv) {

	int server_id, cl_fd, i;
	int childpid;
	struct sockaddr_in g_s_addr;
	struct t_bind g_s_bind;
	struct t_call g_s_call;
/* MY_STUFF_END_3 */

  store_pid();
  srand(time(0) ^ (getpid() << 16));
  printf("Catty v3 by <lcamtuf@coredump.cx> is booting up...\n");
	printf("XTI modifications by <gim@skrzynka.pl>\n");

  setbuffer(stdout,0,0);
  index_kbase();
  load_excuses();


/* MY_STUFF_BEGIN_4 */
	if (prepare_tli(&server_id, &g_s_addr, &g_s_bind, &g_s_call))
		return 1;


	printf(" (G+) Waiting on connection on port: %d\n", DEF_PORT);
	while(1) {
		waitpid(0, NULL, WNOHANG);
		if (t_listen(server_id, &g_s_call) < 0)
		{
			if (t_errno != TSYSERR)
				t_error(" (G!) t_listen() error");
			continue;
		}

		puts(" (G+) Someone's connecting");

		if ((cl_fd = accept_client(server_id, &g_s_call, 1)) < 0)
			continue;

		for (i=0; i<MAXCHILD; i++)
			if (!dzieci[i])
				break;
		if (i != MAXCHILD) {
			if ((childpid = fork()) < 0)
			{
				perror(" (G!) fork()");
				printf(" (G!) failed, rejecting\n");

				g_s_call.addr.maxlen = 0;
				g_s_call.addr.len = 0;
				g_s_call.opt.maxlen = 0;
				g_s_call.opt.len = 0;
				g_s_call.udata.maxlen = 0;
				g_s_call.udata.len = 0;
				g_s_call.sequence = 1;

				t_snddis(cl_fd, &g_s_call);
				t_unbind(cl_fd);
				t_close(cl_fd);
				continue;
			}
			if (childpid == 0)
			{
				t_close(server_id);
				we_all_have_a_part_of_catty_in_ourself(cl_fd);
				return 0;
			}
			printf(" (G+) Connection accepted on slot %d\n", i);
			dzieci[i] = childpid;
		} else {
			g_s_call.addr.maxlen = 0;
			g_s_call.addr.len = 0;
			g_s_call.opt.maxlen = 0;
			g_s_call.opt.len = 0;
			g_s_call.udata.maxlen = 0;
			g_s_call.udata.len = 0;
			g_s_call.sequence = 2;

			t_snddis(cl_fd, &g_s_call);
			t_unbind(cl_fd);
			printf(" (G-) Connection rejected, no free slot\n");
		}
		t_close (cl_fd);
	}
/* MY_STUFF_END_4 */
  fatal("EOF encountered. Talk to the hand.");
  return 0;
}


